from collections.abc import Awaitable, Callable
from typing import Any, Literal, ParamSpec, overload

from bolinette.core import Cache, __user_cache__, meta
from bolinette.core.types import Function

P_Func = ParamSpec("P_Func")


class CommandMeta:
    def __init__(
        self,
        path: str,
        summary: str,
    ):
        self.path = path
        self.summary = summary


class Argument:
    @overload
    def __init__(
        self,
        arg_type: Literal["argument"] = "argument",
        /,
        *,
        default: Any | None = None,
        summary: str | None = None,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        arg_type: Literal["option"],
        shorthand: str | None = None,
        /,
        *,
        default: Any | None = None,
        summary: str | None = None,
    ) -> None:
        ...

    def __init__(
        self,
        /,
        *args: Any,
        **kwargs: Any,
    ):
        self.arg_type: Literal["argument", "option"]
        self.shorthand: str | None
        match args:
            case ():
                self.arg_type = "argument"
                self.shorthand = None
            case ("argument",):
                self.arg_type = "argument"
                self.shorthand = None
            case ("option",):
                self.arg_type = "option"
                self.shorthand = None
            case ("option", str() as shorthand):
                self.arg_type = "option"
                self.shorthand = shorthand
            case _:
                raise TypeError(Argument.__init__, args)
        self.default: Any | None = kwargs.get("default", None)
        self.summary: str | None = kwargs.get("summary", None)


def command(name: str, summary: str, *, cache: Cache | None = None):
    def decorator(func: Callable[P_Func, Awaitable[int | None]]) -> Callable[P_Func, Awaitable[int | None]]:
        meta.set(func, CommandMeta(name, summary))
        (cache or __user_cache__).add(CommandMeta, Function(func))
        return func

    return decorator



import inspect
import sys
from argparse import ArgumentParser
from collections.abc import Awaitable, Callable
from typing import Any, Literal, Protocol

from bolinette.core import Cache, Logger, meta
from bolinette.core.command.command import Argument, CommandMeta
from bolinette.core.exceptions import InitError
from bolinette.core.injection import init_method
from bolinette.core.types import Function, Type


class _SubParsersAction(Protocol):
    def add_parser(self, name: str, *, help: str | None = None) -> ArgumentParser:
        ...


class Parser:
    def __init__(
        self,
        logger: Logger["Parser"],
        cache: Cache,
    ):
        self._cache = cache
        self._logger = logger
        self._commands: dict[str, Function[..., Awaitable[None]]] = {}
        self._sub_commands: dict[str, Any] = {}
        self._parser = ArgumentParser(description="Bolinette Framework")

    @init_method
    def _parse_commands(self) -> None:
        functions = self._cache.get(CommandMeta, hint=Function[..., Awaitable[None]], raises=False)
        commands: dict[str, Function[..., Awaitable[None]]] = {}
        command_tree: dict[str, Any] = {}
        for func in functions:
            _cmd = meta.get(func.func, CommandMeta)
            cur_node = command_tree
            path = _cmd.path.split(" ")
            for elem in path[:-1]:
                if elem not in cur_node:
                    cur_node[elem] = {}
                cur_node = cur_node[elem]
            elem = path[-1]
            if elem in cur_node:
                raise InitError(f"Conflict with '{_cmd.path}' command")
            cur_node[elem] = func
            commands[_cmd.path] = func
        self._commands = commands
        self._build_parsers(command_tree, self._parser.add_subparsers(), [])

    def _build_parsers(
        self,
        command_tree: dict[str, dict[str, Any] | Function[..., Any]],
        sub_parsers: _SubParsersAction,
        path: list[str],
    ) -> None:
        sub_commands: dict[str, Any] = {}
        for name, elem in command_tree.items():
            if isinstance(elem, Function) and meta.has(elem.func, CommandMeta):
                _cmd = meta.get(elem.func, CommandMeta)
                sub_parser = sub_parsers.add_parser(name, help=_cmd.summary)
                annotations = elem.annotations()
                for p_name, param in elem.parameters().items():
                    if p_name not in annotations:
                        continue
                    param_t: Type[Any] = annotations[p_name]
                    for anno in param_t.annotated:
                        if anno is Argument or isinstance(anno, Argument):
                            if anno is Argument:
                                anno = Argument()
                            flags, kwargs = self._create_argument(elem, p_name, param, param_t, anno)
                            sub_parser.add_argument(*flags, **kwargs)
                sub_parser.set_defaults(__blnt_cmd__=_cmd.path)
            elif isinstance(elem, dict):
                sub_parser = sub_parsers.add_parser(name, help=self._build_help(elem, [*path, name]))
                sub_parser.set_defaults(__blnt_path__=name)
                sub_commands[name] = sub_parser
                self._build_parsers(elem, sub_parser.add_subparsers(), [*path, name])
        self._sub_commands = sub_commands

    @staticmethod
    def _create_argument(
        func: Function[..., Any],
        p_name: str,
        param: inspect.Parameter,
        type: Type[Any],
        argument: Argument,
    ) -> tuple[list[str], dict[str, Any]]:
        flags: list[str]
        kwargs: dict[str, Any] = {}
        if param.default != inspect.Signature.empty:
            kwargs["default"] = param.default
        match argument.arg_type:
            case "argument":
                flags = [p_name]
                if type.nullable:
                    raise InitError(f"Command {func}, Argument '{p_name}', A positional argument cannot be nullable")
            case "option":
                flags = [f"--{p_name}"]
                if argument.shorthand:
                    flags.append(f"-{argument.shorthand}")
                if not type.nullable and "default" not in kwargs:
                    kwargs["required"] = True
        if type.cls is str:
            kwargs["type"] = str
        elif type.cls is int:
            kwargs["type"] = int
        elif type.cls is float:
            kwargs["type"] = float
        elif type.cls is Literal:
            if len(type.vars) == 1:
                if type.vars == (True,):
                    kwargs["action"] = "store_true"
                elif type.vars == (False,):
                    kwargs["action"] = "store_false"
                else:
                    kwargs["action"] = "store_const"
                    kwargs["const"] = type.vars[0]
            else:
                if all(isinstance(i, str) for i in type.vars):
                    kwargs["action"] = "store"
                    kwargs["type"] = str
                elif all(isinstance(i, int) for i in type.vars):
                    kwargs["action"] = "store"
                    kwargs["type"] = int
                else:
                    raise InitError(f"Command {func}, Argument '{p_name}', {type} is not a valid argument type")
                kwargs["choices"] = type.vars
        else:
            raise InitError(f"Command {func}, Argument '{p_name}', Type {type} is not allowed as a command argument")
        if argument.summary:
            kwargs["help"] = argument.summary
        return flags, kwargs

    def parse_command(self, args: list[str]) -> tuple[Callable[..., Awaitable[int | None]], dict[str, Any]]:
        parsed = vars(self._parser.parse_args(args))
        if "__blnt_cmd__" in parsed:
            cmd = parsed.pop("__blnt_cmd__")
            if "__blnt_path__" in parsed:
                del parsed["__blnt_path__"]
            return (self._commands[cmd].func, parsed)
        elif "__blnt_path__" in parsed:
            self._logger.error(self._sub_commands[parsed["__blnt_path__"]].format_help())
            sys.exit(1)
        else:
            self._logger.error(self._parser.format_help())
            sys.exit(1)

    @staticmethod
    def _build_help(command_tree: dict[str, Any], path: list[str]):
        commands = [f'"{" ".join([*path, x])}"' for x in command_tree]
        return "Sub-commands: " + ", ".join(commands)



from bolinette.core.objects import GenericMeta as GenericMeta
from bolinette.core.metadata import meta as meta
from bolinette.core.cache import Cache as Cache, __user_cache__ as __user_cache__
from bolinette.core.logger import Logger as Logger, ConsoleColorCode as ConsoleColorCode
from bolinette.core.command.command import command as command
from bolinette.core.extension import core_ext as __blnt_ext__
from bolinette.core.startup import startup as startup
from bolinette.core.bolinette import Bolinette as Bolinette



import collections.abc
import contextlib
import inspect
from types import NoneType, UnionType
from typing import (
    Annotated,
    Any,
    ForwardRef,
    Literal,
    TypeGuard,
    TypeVar,
    Union,
    get_args,
    get_origin,
    get_type_hints,
    override,
)

from bolinette.core import types
from bolinette.core.exceptions import TypingError


class Type[T]:
    __slots__: list[str] = ["cls", "vars", "nullable", "union", "annotated", "_hash"]

    @staticmethod
    def from_instance(__instance: T) -> "Type[T]":
        return Type(type(__instance))

    def __init__(
        self,
        cls: type[T],
        /,
        *,
        lookup: "types.TypeVarLookup[Any] | None" = None,
        raise_on_string: bool = True,
        raise_on_typevar: bool = True,
    ) -> None:
        self.annotated: tuple[Any, ...]
        if get_origin(cls) is Annotated:
            cls, *self.annotated = get_args(cls)
        else:
            self.annotated = ()
        if get_origin(cls) in (Union, UnionType):
            args = get_args(cls)
            self.nullable = None in args or NoneType in args
            args = tuple(a for a in args if a not in (None, NoneType))
            cls, *additional_cls = args
        else:
            additional_cls = ()
            self.nullable = False
        self.cls, self.vars = Type.get_generics(cls, lookup, raise_on_string, raise_on_typevar)
        self.vars = (*self.vars, *map(lambda _: Any, range(len(self.vars), Type.get_param_count(self.cls))))
        self.union = tuple(Type(c) for c in additional_cls)
        self._hash = hash((self.cls, self.vars))

    @override
    def __str__(self) -> str:
        def _format_v(v: Any) -> str:
            if isinstance(v, type):
                return v.__qualname__
            if isinstance(v, TypeVar):
                return f"~{v.__name__}"
            if v is Ellipsis:
                return "..."
            if v is Literal:
                return v.__name__
            if isinstance(v, ForwardRef):
                return f"'{v.__forward_arg__}'"
            return str(v)

        if not self.vars:
            repr_str = _format_v(self.cls)
        else:
            repr_str = f"{_format_v(self.cls)}[{', '.join(map(_format_v, self.vars))}]"

        if self.is_union:
            for t in self.union:
                repr_str += f" | {t}"

        if self.nullable:
            repr_str += " | None"

        return repr_str

    @override
    def __repr__(self) -> str:
        return f"<Type {self}>"

    @override
    def __hash__(self) -> int:
        return self._hash

    @override
    def __eq__(self, __value: object) -> bool:
        return (
            isinstance(__value, Type)
            and __value.cls is self.cls  # pyright: ignore[reportUnknownMemberType]
            and __value.vars == self.vars
        )

    @property
    def init(self):
        return types.Function(self.cls.__init__)

    @property
    def is_union(self) -> bool:
        return len(self.union) > 0

    @property
    def is_any(self) -> bool:
        return self.cls is Any

    def new(self, *args: Any, **kwargs: Any) -> T:
        return self.cls(*args, **kwargs)

    def parameters(self) -> dict[str, inspect.Parameter]:
        return {**inspect.signature(self.cls).parameters}

    def annotations(self, *, lookup: "types.TypeVarLookup[Any] | None" = None) -> "dict[str, Type[Any]]":
        return self._get_recursive_annotations(self.cls, lookup)

    def isinstance(self, instance: Any) -> TypeGuard[T]:
        return isinstance(instance, self.cls)

    @staticmethod
    def _get_recursive_annotations(
        _cls: type[Any], lookup: "types.TypeVarLookup[Any] | None"
    ) -> "dict[str, Type[Any]]":
        annotations: dict[str, Type[Any]] = {}
        try:
            for base in _cls.__bases__:
                annotations |= Type._get_recursive_annotations(base, lookup)
            hints: dict[str, type[Any]] = get_type_hints(_cls)
            for attr_name, hint in hints.items():
                annotations[attr_name] = Type(hint, lookup=lookup)
        except (AttributeError, TypeError, NameError):
            return annotations
        return annotations

    @staticmethod
    def get_generics(
        _cls: type[T],
        lookup: "types.TypeVarLookup[Any] | None",
        raise_on_string: bool,
        raise_on_typevar: bool,
    ) -> tuple[type[T], tuple[Any, ...]]:
        if origin := get_origin(_cls):
            type_vars: tuple[Any, ...] = ()
            for arg in get_args(_cls):
                arg: Any
                if isinstance(arg, ForwardRef) and raise_on_string:
                    raise TypingError(
                        f"Generic parameter '{arg.__forward_arg__}' cannot be a string", cls=origin.__qualname__
                    )
                if isinstance(arg, TypeVar):
                    if lookup is not None:
                        if arg in lookup:
                            arg = lookup[arg]
                        elif raise_on_typevar:
                            raise TypingError(
                                f"TypeVar ~{arg.__name__} could not be found in lookup", cls=origin.__qualname__
                            )
                    elif raise_on_typevar:
                        raise TypingError(
                            f"Generic parameter ~{arg.__name__} cannot be a TypeVar", cls=origin.__qualname__
                        )
                if isinstance(arg, list):
                    arg = tuple(*arg)  # pyright: ignore[reportUnknownArgumentType]
                type_vars = (*type_vars, arg)
            return origin, type_vars
        return _cls, ()

    @staticmethod
    def get_param_count(cls_: type[Any]) -> int:
        if cls_ in _BUILTIN_PARAM_COUNT:
            return _BUILTIN_PARAM_COUNT[cls_]
        if hasattr(cls_, "__parameters__"):
            return len(cls_.__parameters__)  # pyright: ignore
        return 0


_BUILTIN_PARAM_COUNT: dict[type[Any], int] = {
    collections.abc.Hashable: 0,
    collections.abc.Awaitable: 1,  # pyright: ignore
    collections.abc.Coroutine: 3,  # pyright: ignore
    collections.abc.AsyncIterable: 1,  # pyright: ignore
    collections.abc.AsyncIterator: 1,  # pyright: ignore
    collections.abc.Iterable: 1,  # pyright: ignore
    collections.abc.Iterator: 1,  # pyright: ignore
    collections.abc.Reversible: 1,  # pyright: ignore
    collections.abc.Sized: 0,
    collections.abc.Container: 1,  # pyright: ignore
    collections.abc.Collection: 1,  # pyright: ignore
    collections.abc.Set: 1,  # pyright: ignore
    collections.abc.MutableSet: 1,  # pyright: ignore
    collections.abc.Mapping: 2,  # pyright: ignore
    collections.abc.MutableMapping: 2,  # pyright: ignore
    collections.abc.Sequence: 1,  # pyright: ignore
    collections.abc.MutableSequence: 1,  # pyright: ignore
    list: 1,
    collections.deque: 1,  # pyright: ignore
    set: 1,
    frozenset: 1,
    collections.abc.MappingView: 1,
    collections.abc.KeysView: 1,  # pyright: ignore
    collections.abc.ItemsView: 2,  # pyright: ignore
    collections.abc.ValuesView: 1,  # pyright: ignore
    contextlib.AbstractContextManager: 1,  # pyright: ignore
    contextlib.AbstractAsyncContextManager: 1,  # pyright: ignore
    dict: 2,
    collections.defaultdict: 2,  # pyright: ignore
    collections.OrderedDict: 2,  # pyright: ignore
    collections.Counter: 1,  # pyright: ignore
    collections.ChainMap: 2,  # pyright: ignore
    collections.abc.Generator: 3,  # pyright: ignore
    collections.abc.AsyncGenerator: 2,  # pyright: ignore
    type: 1,
}



from collections.abc import Iterable
from typing import Any, Literal, overload

from bolinette.core.types import Type


class InjectionContext:
    __slots__ = ("_instances",)

    def __init__(self) -> None:
        self._instances: dict[Type[Any], Any] = {}

    def has_instance(self, t: Type[Any]) -> bool:
        return t in self._instances

    def set_instance[InstanceT](self, t: Type[InstanceT], instance: InstanceT) -> None:
        self._instances[t] = instance

    def get_instance[InstanceT](self, t: Type[InstanceT]) -> InstanceT:
        return self._instances[t]

    @overload
    def get_instances(self) -> Iterable[Any]:
        ...

    @overload
    def get_instances(self, *, with_types: Literal[True]) -> Iterable[tuple[type[Any], Any]]:
        ...

    def get_instances(self, **kwargs: Any) -> Iterable[tuple[type[Any], Any]] | Iterable[Any]:
        if "with_types" in kwargs:
            return [(t, i) for t, i in self._instances.items()]
        return [i for i in self._instances.values()]



from collections.abc import Callable

from bolinette.core import Cache, __user_cache__

STARTUP_CACHE_KEY = object()


def startup[**P, T](
    func: Callable[P, T],
    *,
    cache: Cache | None = None,
) -> Callable[P, T]:
    (cache or __user_cache__).add(STARTUP_CACHE_KEY, func)
    return func



import asyncio
import sys
from collections.abc import Callable
from typing import Any, NoReturn

from bolinette import core
from bolinette.core import Logger, __user_cache__, meta
from bolinette.core.cache import Cache
from bolinette.core.command import Parser
from bolinette.core.environment import Environment
from bolinette.core.exceptions import InitError
from bolinette.core.extension import Extension, ExtensionModule
from bolinette.core.injection import Injection, require
from bolinette.core.startup import STARTUP_CACHE_KEY


class Bolinette:
    def __init__(self) -> None:
        self._cache = Cache()
        self._initialized = False
        self._extensions: list[Extension] = [core.__blnt_ext__]
        self._inject: Injection
        self._env: Environment
        self._logger: Logger[Bolinette]

    def use_extension[ExtT: Extension](self, ext: ExtensionModule[ExtT] | ExtT) -> ExtT:
        if self._initialized:
            raise InitError("Cannot use extension after Bolinette startup")
        if not isinstance(ext, Extension):
            ext = ext.__blnt_ext__
        if ext in self._extensions:
            raise InitError(f"Bolinette extension {ext.name} is already in use")
        self._extensions.append(ext)
        for dep_ext in ext.dependencies:
            if dep_ext not in self._extensions:
                self._extensions.append(dep_ext)
        return ext

    @require(Parser)
    def _parser(self):
        ...

    async def _run_startup_funcs(self) -> None:
        funcs = self._cache.get(STARTUP_CACHE_KEY, hint=Callable[..., Any], raises=False)
        async with self._inject.get_async_scoped_session() as scoped_inject:
            for func in funcs:
                if asyncio.iscoroutine(res := scoped_inject.call(func)):
                    await res

    async def startup(self) -> None:
        if self._initialized:
            raise InitError("Bolinette has already been initialized")
        self._extensions = Extension.sort_extensions(self._extensions)
        for ext in self._extensions:
            ext.add_cached(self._cache)

        self._cache |= __user_cache__

        self._inject = Injection(self._cache)
        meta.set(self, self._inject)
        self._env = self._inject.require(Environment)
        self._cache.debug = self._env.config["core"]["debug"]

        self._inject.add(Bolinette, "singleton", instance=self)
        self._inject.__hook_proxies__(self)

        await self._run_startup_funcs()

        self._logger = self._inject.require(Logger[Bolinette])
        self._logger.info(f"Loaded Bolinette with extensions: {', '.join(e.name for e in self._extensions)}")
        self._initialized = True

    async def exec_args(self, args: list[str]) -> NoReturn:
        if not self._initialized:
            await self.startup()
        cmd, cmd_args = self._parser.parse_command(args)
        result = await self._inject.call(cmd, named_args=cmd_args)
        sys.exit(result or 0)
